@@ -1,7 +1,7 @@
 /*
  * utils.c - Misc utilities
  *
- * Copyright (C) 2013 - 2017, Max Lv <max.c.lv@gmail.com>
+ * Copyright (C) 2013 - 2018, Max Lv <max.c.lv@gmail.com>
  *
  * This file is part of the shadowsocks-libev.
  *
@@ -25,18 +25,21 @@
 #endif
 
 #include <stdlib.h>
-#include <unistd.h>
 #include <string.h>
-#include <errno.h>
 #include <ctype.h>
+#ifndef __MINGW32__
+#include <unistd.h>
+#include <errno.h>
 #include <pwd.h>
 #include <grp.h>
+#endif
 
 #include <sys/types.h>
 #include <sys/stat.h>
 
 #include <sodium.h>
 
+#include "crypto.h"
 #include "utils.h"
 
 #ifdef HAVE_SETRLIMIT
@@ -54,12 +57,14 @@ FILE *logfile;
 int use_syslog = 0;
 #endif
 
+#ifndef __MINGW32__
 void
 ERROR(const char *s)
 {
     char *msg = strerror(errno);
     LOGE("%s: %s", s, msg);
 }
+#endif
 
 int use_tty = 1;
 
@@ -86,7 +91,8 @@ ss_itoa(int i)
 }
 
 int
-ss_isnumeric(const char *s) {
+ss_isnumeric(const char *s)
+{
     if (!s || !*s)
         return 0;
     while (isdigit((unsigned char)*s))
@@ -100,6 +106,7 @@ ss_isnumeric(const char *s) {
 int
 run_as(const char *user)
 {
+#ifndef __MINGW32__
     if (user[0]) {
         /* Convert user to a long integer if it is a non-negative number.
          * -1 means it is a user name. */
@@ -125,7 +132,7 @@ run_as(const char *user)
              * which returns its result in a statically allocated buffer and
              * cannot be considered thread safe. */
             err = uid >= 0 ? getpwuid_r((uid_t)uid, &pwdbuf, buf, buflen, &pwd)
-                : getpwnam_r(user, &pwdbuf, buf, buflen, &pwd);
+                  : getpwnam_r(user, &pwdbuf, buf, buflen, &pwd);
 
             if (err == 0 && pwd) {
                 /* setgid first, because we may not be allowed to do it anymore after setuid */
@@ -153,7 +160,7 @@ run_as(const char *user)
             } else if (err != ERANGE) {
                 if (err) {
                     LOGE("run_as user '%s' could not be found: %s", user,
-                            strerror(err));
+                         strerror(err));
                 } else {
                     LOGE("run_as user '%s' could not be found.", user);
                 }
@@ -172,7 +179,7 @@ run_as(const char *user)
         /* No getpwnam_r() :-(  We'll use getpwnam() and hope for the best. */
         struct passwd *pwd;
 
-        if (!(pwd = uid >=0 ? getpwuid((uid_t)uid) : getpwnam(user))) {
+        if (!(pwd = uid >= 0 ? getpwuid((uid_t)uid) : getpwnam(user))) {
             LOGE("run_as user %s could not be found.", user);
             return 0;
         }
@@ -193,6 +200,9 @@ run_as(const char *user)
         }
 #endif
     }
+#else
+    LOGE("run_as(): not implemented in MinGW port");
+#endif
 
     return 1;
 }
@@ -300,6 +310,10 @@ usage()
         "                                  camellia-256-cfb, bf-cfb,\n");
     printf(
         "                                  chacha20-ietf-poly1305,\n");
+#ifdef FS_HAVE_XCHACHA20IETF
+    printf(
+        "                                  xchacha20-ietf-poly1305,\n");
+#endif
     printf(
         "                                  salsa20, chacha20 and chacha20-ietf.\n");
     printf(
@@ -354,6 +368,8 @@ usage()
         "       [--fast-open]              Enable TCP fast open.\n");
     printf(
         "                                  with Linux kernel > 3.7.0.\n");
+#endif
+#if defined(MODULE_REMOTE) || defined(MODULE_LOCAL)
     printf(
         "       [--acl <acl_file>]         Path to ACL (Access Control List).\n");
 #endif
@@ -372,6 +388,8 @@ usage()
         "       [--mptcp]                  Enable Multipath TCP on MPTCP Kernel.\n");
 #endif
 #ifndef MODULE_MANAGER
+    printf(
+        "       [--no-delay]               Enable TCP_NODELAY.\n");
     printf(
         "       [--key <key_in_base64>]    Key of your remote server.\n");
 #endif
@@ -390,6 +408,7 @@ usage()
 void
 daemonize(const char *path)
 {
+#ifndef __MINGW32__
     /* Our process ID and Session ID */
     pid_t pid, sid;
 
@@ -434,6 +453,9 @@ daemonize(const char *path)
     close(STDIN_FILENO);
     close(STDOUT_FILENO);
     close(STDERR_FILENO);
+#else
+    LOGE("daemonize(): not implemented in MinGW port");
+#endif
 }
 
 #ifdef HAVE_SETRLIMIT
@@ -464,3 +486,32 @@ set_nofile(int nofile)
 }
 
 #endif
+
+char *
+get_default_conf(void)
+{
+#ifndef __MINGW32__
+    static char sysconf[] = "/etc/shadowsocks-libev/config.json";
+    static char userconf[PATH_MAX] = { 0 };
+    char *conf_home;
+
+    conf_home = getenv("XDG_CONFIG_HOME");
+
+    if (!conf_home) {
+        strcpy(userconf, getenv("HOME"));
+        strcat(userconf, "/.config/shadowsocks-libev/config.json");
+    } else {
+        strcpy(userconf, conf_home);
+        strcat(userconf, "/shadowsocks-libev/config.json");
+    }
+
+    // Check if the user-specific config exists.
+    if (access(userconf, F_OK) != -1)
+        return userconf;
+
+    // If not, fall back to the system-wide config.
+    return sysconf;
+#else
+    return "config.json";
+#endif
+}
